#!/usr/bin/env python

"""Network parsing utilities.

Currently developed for parsing, analyzing and plotting network
output files in newick-like format produced by the SNAQ program in
PhyloNetworks.

Toytree default is to draw networks in "majortree" mode.

Extended newick network format
------------------------------
Basic network, e.g., used by dendroscope.
>>> ((r3,(r2,(r1, #HX)),(r0)#HX);
This indicates an admix edges from r1 -> r0. The #HX node is the
source, and the #HX node label is the destination. This example shows
only the edge, but not its proportions, or edge lengths indicating
where the admix edge starts or stops at each end.

With admixture magnitudes (termed gamma values under some models).
>>> ((r3,(r2, (r1, #HX:::0.1)),(r0)#HX);

With admixture magnitudes and edge lengths.
>>>

This is what it looks like when more than one admix from a node:
>>> ((A,#H1),B)
>>> (((A,#H1),H2),B)

This is what it looks like when more than one admix to a node:
>>> (A)#H1
>>> (A)#H1#H2  (???)

Examples
---------
>>> NET = "(r1,(r2,(r3,(r4,#H6):10.0):1.949048825686914):10.0,(r0)#H6);""
>>> toytree.utils.network.parse_networks(NET)
"""

from typing import Union, Tuple, List
from pathlib import Path
import re
from loguru import logger
import toytree
from toytree.io.src.utils import replace_whitespace

logger = logger.bind(name="toytree")


class Network:
    """Child class of ToyTree that stores and draws network edges.

    A network object can be used to draw networks, to calculate or
    report statistics on a network, and to extract minor or major
    trees from a network (as ToyTree objects).

    Algorithm to traverse nodes of a network in order... could still
    use the same traversals that trace the major or minor tree to visit
    each node, but another traversal may be available that would also
    visit nodes connected by an admix edge sooner...
    """
    def __init__(self, network: Union[str, Path]):
        self.network = network
        # tree, admix = parse_network(net, disconnect=disconnect)
        # super().__init__(tree, *args, **kwargs)
        # self.admix = admix

    # def ... , disconnect: bool = True, **kwargs):

    def add_edge(self) -> toytree.ToyTree:
        """Return modified Network with a new degree-3 edge added."""

    def remove_edge(self) -> toytree.ToyTree:
        """Return modified Network with a degree-3 edge removed."""

    def get_major_tree(self) -> toytree.ToyTree:
        """Return the major tree..."""

    def get_admixture_edges(self) -> Tuple:
        """Return Tuple with list of admixture edges..."""

    def write(self):
        """Write network in extended format..."""

    def root(self):
        """raise rootMisMatchError if direction doesn't work..."""


class NetworkToMajorTree:
    """Class with functions to parse (ToyTree, admix_edges) from net.

    ...
    """
    def __init__(self, net: Union[str, Path], disconnect=True):
        self.net: str = self._get_network_as_newick_str(net)
        """: Input network as newick string (parsed from str or Path)."""
        self.tree: toytree.ToyTree = None
        """: The major tree."""
        self.admix: List[Tuple] = []
        """: List of admixture edges."""
        self.disconnect: bool = disconnect
        """: ..."""

    def _get_network_as_newick_str(self, net: Union[str, Path]) -> str:
        """Parse network newick from file or '[network];...details' string."""
        # if net is a file then read the first line
        if ";" not in net:
            if not net.exists():
                raise IOError("input type is not a file or newick (no ; ending).")
            with open(net, 'rt', encoding="utf-8") as infile:
                net = infile.readline()
        else:
            # trim off loglik and anything after it (TODO: keep loglik)
            if ";" in net:
                net = net.split(";")[0] + ';'
        # strip whitespace
        net = replace_whitespace(net)
        # net = WHITE_SPACE.sub("", net)
        return net

    def _set_tree_with_admix_as_node_labels(self) -> toytree.ToyTree:
        """Return a ToyTree parsed from network newick with gamma nodes.

        Admix nodes are indicated by #[name]:[dist]::[gamma] where dist
        will be absent for tip nodes but present for internal nodes. We
        modify this to be #[name-with-gamma]:dist so we can still use
        the same tree parser and then modify the tree afterwards to
        extract the major tree and minor admix edges.
        """
        string = self.net
        # tip nodes replace L:::G -> L-G
        re0 = re.compile(r"(#\w+):::([\d.]+)")
        # internal nodes replace L:D::G -> L-G:D
        re1 = re.compile(r"(#\w+):([\d.]+)::([\d.]+)")
        string = re0.sub(r"\1-gamma-\2", string)
        string = re1.sub(r"\1-gamma-\3:\2", string)
        self.tree = toytree.tree(string)

    def _pseudo_unroot(self) -> toytree.ToyTree:
        """Return unrooted tree re-ordered for display purposes.

        Tree is rooted tree on an edge that does not split a ...
        """
        # store a list of all nodes the tree *could* be rooted on.
        opt = self.tree.get_nodes()
        for node in self.tree.traverse("postorder"):
            if node.name.startswith("#H"):
                for desc in node.get_descendants():
                    if desc in opt:
                        opt.remove(desc)
        logger.trace(f"can root on {opt}")
        return self.tree.root(opt[0]).mod.ladderize().mod.unroot()

    def get_major_tree_and_admix_edges(self) -> Tuple:
        """..."""
        self._set_tree_with_admix_as_node_labels()
        self.tree._draw_browser(ts='s', layout='unr', width=700, node_labels="name", node_colors="pink")
        self.tree = self._pseudo_unroot()
        self.tree._draw_browser(ts='s', width=500, node_labels="idx", node_colors="pink", tmpdir="~")

        # restart traversal after each pair of nodes is fixed.
        while 1:
            try:
                hnodes = self.tree.get_nodes("~#H*")
                hnodes = sorted(hnodes, key=lambda x: x.name)
            except ValueError:
                self.tree.mod.remove_unary_nodes(inplace=True)
                logger.debug(f"finished: {self.admix}")
                return self.tree, self.admix

            # order as [dest, src] i.e., [node-label, tip-node] and get gamma
            if hnodes[0].is_leaf():
                hnodes = [hnodes[1], hnodes[0]]
            if "-gamma-" in hnodes[1].name:
                prop = hnodes[1].name.rsplit("-gamma-", 1)[1]
            else:
                prop = 0.5
            logger.trace(f"hnodes: {hnodes}")

            # remove the tip node representing a src
            logger.debug(f"deleting minor node: {hnodes[1].name} {hnodes[1].idx}")
            desc1 = tuple(hnodes[1]._up.get_leaf_names())
            hnodes[1]._delete()
            desc1 = [i for i in desc1 if not i.startswith("#H")]
            logger.trace(f"desc1: {desc1}")

            # remove internal node representing admix destination
            logger.debug(f"deleting major node: {hnodes[0].name} {hnodes[0].idx}")
            desc0 = tuple(hnodes[0].get_leaf_names())
            desc0 = [i for i in desc0 if not i.startswith("#H")]
            hnodes[0]._delete()
            logger.trace(f"desc0: {desc0}")

            self.tree._update()
            self.admix.append((desc0, desc1, 0.5, {}, f"{float(prop):.3g}"))


def parse_network(net: Union[str, Path], disconnect=True):
    """Parse a network file to extract a major topology and admix dict.

    This leaves the hybrid nodes in the tree and labels each with
    .name="H{int}" and .gamma={float}.
    """
    # if net is a file then read the first line
    if ";" not in net:
        if not net.exists():
            raise IOError("input type is not a file or newick (no ; ending).")
        with open(net, 'rt', encoding="utf-8") as infile:
            net = infile.readline()
    else:
        # trim off loglik and anything after it (TODO: keep loglik)
        if ";" in net:
            net = net.split(";")[0] + ';'

    # sub :xxx:: to be ::: b/c I don't care about admix edge bls FOR NOW.
    net = re.sub(r":\d.\w*::", ":::", net)
    print(net)

    # change H nodes to be internal node labels rather than tip nodes.
    while ",#" in net:
        pre, post = net.split(",#", 1)
        npre, npost = post.split(")", 1)
        newpre = npre.split(":")[0] + "-" + npre.split(":")[-1]
        net = pre + ")#" + newpre + npost
    net = net.replace(":::", "-")
    print(net)

    # parse cleaned newick and set empty gamma on all nodes
    net = toytree.tree(net)#, tree_format=1)
    print(net.get_node_data())

    # store admix data
    admix = {}

    # if not rooted choose any non-H root
    if not net.is_rooted():
        net = net.root(
            [i for i in net.get_tip_labels() if not i.startswith("#H")][0]
        )

    # Traverse tree to find hybrid nodes. If a hybrid node is labeled as a
    # distinct branch in the tree then it is dropped from the tree and
    for node in net.traverse("postorder"):

        # find hybrid nodes as internal nchild=1, or external with H in name
        if (len(node.children) == 1) or node.name.startswith("#H"):

            # assign name and gamma to hybrid nodes
            print(node.name)
            aname, aprop = node.name.split("-")
            aname = aname.lstrip("#")
            node.name = aname

            # assign hybrid to closest nodes up and down from edge
            # node.children[0].hybrid = int(aname[1:])
            # node.gamma = round(float(aprop), 3)
            # node.up.hybrid = int(aname[1:])

            # if root is a hybrid edge (ugh)
            if node.up is None:
                small, big = sorted(node.children, key=len)
                root = toytree.Node(name='root')
                node.children = [small]
                small.up = node
                node.up = root
                big.up = root
                root.children = [node, big]
                net.treenode = root

            # disconnect node by connecting children to parent
            if disconnect:

                # if tip is a hybrid
                if not node.children:
                    # get sister node
                    sister = [i for i in node.up.children if i != node][0]

                    # connect sister to gparent
                    sister.up = node.up.up
                    node.up.up._remove_child(node.up)
                    node.up.up._add_child(sister)

                # if hybrid is internal
                else:
                    node.up._remove_child(node)
                    for child in node.children:
                        child._up = node.up
                        node.up._add_child(child)

            # store admix data by descendants but remove hybrid tips
            desc = node.get_leaf_names()
            if aname in desc:
                desc = [i for i in node.up.get_leaf_names() if i != aname]
            desc = [i for i in desc if not i.startswith("#H")]

            # put this node into admix
            if aname not in admix:
                admix[aname] = (desc, aprop)

            # matching edge in admix, no arrange into correct order by minor
            else:
                # this is the minor edge
                if aprop < admix[aname][1]:
                    admix[aname] = (
                        admix[aname][0],
                        desc,
                        0.5,
                        {},
                        str(round(float(aprop), 3)),
                    )

                # this is the major edge
                else:
                    admix[aname] = (
                        desc,
                        admix[aname][0],
                        0.5,
                        {},
                        str(round(float(admix[aname][1]), 3)),
                    )

    # update coords needed if node disconnection is turned back on.
    # net._coords.update()
    net._update()
    net = net.mod.ladderize()
    return net, admix

# def test3A():
#     net = "(((r0,#H1:::0.9),r2),(r3,(r4,#H1:::0.1)));"
#     return parse_network(net)

def test3A():
    """No gamma values."""
    net = "(((r0,#H1),r2),(((r3,r4))#H1));"
    parse = NetworkToMajorTree(net)
    tree, admix = parse.get_major_tree_and_admix_edges()
    return tree, admix

def test3B():
    """Major tree has higher gamma value.
    >>> ((r0,r2),(r3,r4));  [(r0,(r3, r4), 0.5)]
    """
    # net = "(((r0,r1),#HX),(((r2,r3))#HX,r4),r5);"
    # net = "(C,D,((O,(E,#H7:::0.196):0.314):0.664,(((A1,A2))#H7:::0.804,B):10.0):10.0);"
    # net = "(C,D,((O,(E,#H7)),(B,(A)#H7)));"
    # net = "(C,D,((O,(E,#H7:::0.49)),(B,(A)#H7:::0.51)));"
    # net = "(C,D,((O,(E,#H7:::0.51)),(B,(A)#H7:::0.49)));"
    net = "((r5,(r4,(r3,((r2,(r0,r1):4.268188213461936):2.9023407392778653)#H9:10.0::0.7093477022612464):3.2038798063864555):10.0):10.0,r6,(r7,#H9:0.0::0.2906522977387535):10.0);"
    # net = "(C,D,((O,(E,#H7:::0.195):0.313):0.664,(B,(A)#H7:::0.804):10.0):10.0);"
    parse = NetworkToMajorTree(net)
    tree, admix = parse.get_major_tree_and_admix_edges()
    return tree, admix

def test3C():
    """...
    >>> ((r0,r2),(r3,r4));  [((r3, r4),r0, 0.5)]
    """
    NET = "(((r0,#H1:::0.9),r2),(r3,r4)#H1:::0.1);"

def test_am_2():
    net = """(quitensis,caudatus,((hybridus,((cruentus,((((retroflexus,wrightii):3.177882128684601,powellii):1.0960348982483923,acutilobus):1.196100087172878,((((((((tuberculatus,arenicola):0.6109501450803072,pumilus):0.4760189778806285,cannabinus):2.412166114595602,(((albus,blitoides):3.721391761165372,(viridis,(blitum,tricolor):1.0147162921429302):3.1242417846101658):0.8709001073072796,(fimbriatus)#H24:::0.5252749295119973):1.8571548128366557):2.707582965691606,#H24:::0.4747250704880027):1.5826753575568118,(((palmeri,watsonii):0.9398829701307351,spinosus):1.724020571200692)#H26:5.622450726846983::0.7964515993969742):3.198826088104374,(dubius,#H26:0.0005151894132288451::0.20354840060302581):9.999954155584865):0.6322904226943913)#H25:0.49722570736992866::0.7233188613740416):2.5040114155558104):0.47630091121000095,hypochondriacus):0.754444529685924):0.7024345634254702,#H25:2.415867402213081::0.2766811386259585):3.554563433506071);"""
    parse = NetworkToMajorTree(net)
    tree, admix = parse.get_major_tree_and_admix_edges()
    return tree, admix
    # parse = NetworkToMajorTree(NET)
    # return parse
    # return parse_network(NET, disconnect=0)

def test_interactive_3():
    NET = "(r1,(r2,(r3,(r4,#H6):10.0):1.949048825686914):10.0,(r0)#H6);"
    NET = "((r3,(r2,(r1, #HX))),(r0)#HX);"
    parse = NetworkParser(NET)
    # parse_network(NET)


if __name__ == "__main__":

    toytree.set_log_level("TRACE")
    # t0, a0 = test3B()
    NET = "(r1,(r2,(r3,(r4,#H6):10.0):1.949048825686914):10.0,(r0)#H6);"
    # parse_network(NET)
    parser = NetworkToMajorTree(NET)
    tree, edges = parser.get_major_tree_and_admix_edges()
    tree._draw_browser(tmpdir="~", admixture_edges=edges)

    # t0, a0 = test_am_2()
    # # t0 = t0.root("fimbriatus")
    # t0._draw_browser(
    #     ts='s',
    #     width=500, height=800,
    #     node_labels="idx",
    #     use_edge_lengths=True,
    #     admixture_edges=a0,
    # )



#!/usr/bin/env python

"""Metrics to calculate phylogenetic signal in trait values.

Blomberg's K is used to quantify phylogenetic signal relative in trait
evolution relative to a Brownian motion model. Values of K>1 indicate
samples are less similar than expected, whereas K<1 indicates that they
are more similar than expected. Permutations can be used to perform
a significance test.

Example
-------
>>> tree = toytree.rtree.unittree(ntips=24, seed=123)
>>> trait = tree.pcm.simulate_continuous_brownian([1.0], tips_only=True, seed=123)
>>> kstat = phylogenetic_signal_k(tree=tree, data=trait, test=True)
>>> # {'K': 0.9857885, 'P-value': 0.002, 'permutations': 1000}

References
----------
The original description of the K statistic:  
__Blomberg, S. P., T. Garland Jr., and A. R. Ives (2003) Testing for
phylogenetic signal in comparative data: Behavioral traits are more
labile. Evolution, *57*, 717-745.__

Extension to conduct hypothesis tests and incorporate sampling error:  
__Ives, A. R., P. E. Midford, and T. Garland Jr. (2007) Within-species
variation and measurement error in phylogenetic comparative biology.
Systematic Biology, *56*, 252-270.__

Extension to multivariate measure of K:  
__Philipp Mitteroecker, Michael L Collyer, Dean C Adams, Exploring Phylogenetic
Signal in Multivariate Phenotypes by Maximizing Blomberg’s K, Systematic 
Biology, 2024;, syae035, https://doi.org/10.1093/sysbio/syae035__

__Adams, D. C. (2014). A generalized K statistic for estimating 
phylogenetic signal from shape and other high-dimensional multivariate
data. Systematic biology, 63(5), 685-697.__

"""

from typing import Union, Sequence, TypeAlias
import numpy as np
import pandas as pd
from toytree.core import ToyTree
from toytree.pcm import get_vcv_matrix_from_tree
from scipy.optimize import minimize_scalar
from loguru import logger


logger = logger.bind(name="toytree")
feature: TypeAlias = Union[str, Sequence[float], pd.Series, pd.DataFrame]
__all__ = ["phylogenetic_signal_k"]


def _validate_features(x: feature, max_dim: int, size: int) -> np.ndarray:
    """Validate data has correct dimensions and size."""
    # if DataFrame w/ only 1 column convert to Series
    if isinstance(x, pd.DataFrame):
        if x.shape[1] == 1:
            x = x.iloc[:, 0]
    # force to array
    x = np.asarray(x)
    # check dimensions and size
    assert x.ndim <= max_dim, f"feature ndim ({x.ndim}) exceeds max allowed ndim ({max_dim})."
    assert x.shape[0] == size, "feature cannot exceed ntips"
    return x


def phylogenetic_signal_k(
    tree: ToyTree,
    data: Union[str, Sequence[float]],
    error: Union[str, Sequence[float]] = None,
    test: bool = False,
    permutations: int = 1000,
) -> dict[str, float]:
    """Return Blomberg's K measurement of phylogenetic signal.

    The K statistic is standardized by the expectation under the
    Brownian motion model of evolution so that K<1 indicates that
    relatives resemble each other less than expected under the model
    (perhaps due to adaptive evolution); whereas K>1 indicates that
    close relatives are more similar than expected under the model.

    Parameters
    ----------
    tree: ToyTree
        A tree with edge lengths.
    data: str | Sequence[float]
        Continuous trait values. 
    error: str | Sequence[float]
        Optional standard errors measured on trait values. 
    test: bool
        Perform permutation test for significance.
    permutations: int
        Number of permutations to perform if testing significance.

    Returns
    -------
    dict[str, float]
        A dict with keys: ["K", "P-value", "permutations"]. It can also
        include additional items if standard errors are included which
        will estimate the variance parameter "sig2".

    Example
    -------
    >>> tree = toytree.rtree.unittree(ntips=10, seed=123, treeheight=2.0)
    >>> data = tree.pcm.simulate_continuous_brownian(1.0, tips_only=True)
    >>> tree.pcm.phylogenetic_signal_k(tree, data)
    >>> # {"K": ..., "P-value": ..., ...}
    """
    if error is None:
        return _phylogenetic_signal_k(tree, data, None, test, permutations)
    else:
        logger.warning("IN DEVELOPMENT")
        return _phylogenetic_signal_k_with_se(tree, data, error, test, permutations)



def _phylogenetic_signal_k(
    tree: ToyTree,
    data: Union[str, Sequence[float]],
    error: Union[str, Sequence[float]] = None,
    test: bool = False,
    permutations: int = 1000,
) -> dict[str, float]:
    """Return Blomberg's K measurement of phylogenetic signal.

    See docstring in `phylogenetic_signal_k`.
    """
    # get ntips and the variance-covariance matrix from tree
    ntips = tree.ntips
    V = get_vcv_matrix_from_tree(tree)

    # [optional] get data as features from the tree
    if isinstance(data, str):
        data = tree.get_node_data(data)[:ntips]
    if isinstance(error, str):
        error = tree.get_node_data(error)[:ntips]
    if error is None:
        error = np.repeat(np.nan, ntips)

    # validate proper trait format returned as float array
    x = _validate_features(data, max_dim=1, size=tree.ntips)    

    # calculate K statistic
    kstat = _calculate_k(x, V)

    # [optional] permutation test
    if test:
        pval = _permutation_test_k(permutations, x, V, kstat)
        return {"K": kstat, "P-value": pval, "permutations": permutations}
    return {"K": kstat, "P-value": np.nan, "permutations": 0}


def _calculate_k(x, V, IV = None) -> float:
    """Return K statistic calculated for data x and variance-covariance
    matrix V."""
    # compute PGLS mean (root state)
    n = x.size
    IV = IV if IV is not None else np.linalg.inv(V)
    a = np.sum(IV @ x) / np.sum(IV)

    # calculate K statistic
    num = ((x - a).T @ (x - a) / ((x - a).T @ IV @ (x - a)))
    dnm = ((np.sum(V.diagonal()) - n / np.sum(IV)) / (n - 1))
    return num / dnm


def _permutation_test_k(size: int, x: np.ndarray, V: np.ndarray, k: float):
    """Return p-value from permutations as a test statistic. 
    """
    kstats = np.zeros(size)
    rng = np.random.default_rng()
    for i in range(size):
        _x = rng.choice(x, size=x.size, replace=False)
        kstats[i] = _calculate_k(_x, V)

    # the proportion of permutations w/ k_ > k
    return np.sum(kstats >= k) / kstats.size


def _calculate_k_with_se(x: np.ndarray, V: np.ndarray, error: np.ndarray) -> tuple[float, float]:
    """
    """
    # start using no error vcv
    IV = np.linalg.inv(V)
    a = np.sum(IV @ x) / np.sum(IV)
    n = x.size
    E = np.diag(error ** 2)    

    # constrain optimization by setting a max on sigma2
    term = x - a
    max_sig2 = (term.T @ IV @ term) / n

    # maximum likelihood model fitting
    estimate = minimize_scalar(
        _likelihood,
        args=(V, E, x),
        bounds=(1e-6, max_sig2),
        method="bounded",
    )
    model_fit = {
        "optimum": estimate.x,
        "LogLik": -estimate.fun,
        "convergence": estimate.success,
    }
    # logger.debug(f"\n{pd.Series(model_fit)}")

    # get rate parameter
    sig2 = model_fit["optimum"] * (n / (n - 1))

    # get VCV w/ rate scalar
    Ve = sig2 * V + E

    # calculate K using optimized Ve
    IVe = np.linalg.inv(Ve)
    a = np.sum(IVe @ x) / np.sum(IVe)
    num = ((x - a).T @ (x - a) / ((x - a).T @ IVe @ (x - a)))
    dnm = ((np.sum(Ve.diagonal()) - n / np.sum(IVe)) / (n - 1))
    return num / dnm, sig2, model_fit["LogLik"]


def _permutation_test_k_with_se(size: int, x: np.ndarray, V: np.ndarray, error: np.ndarray, kstat: float):
    """Return p-value from permutations as a test statistic. 
    """
    kstats = np.zeros(size)
    rng = np.random.default_rng()
    for i in range(size):
        order = rng.choice(range(x.size), size=x.size, replace=False)
        _x = x[order]
        _e = error[order]
        _E = np.diag(_e ** 2)
        kstats[i], _, _ = _calculate_k_with_se(_x, V, _E)

    # the proportion of permutations w/ k_ > k
    return np.sum(kstats >= kstat) / kstats.size


def _phylogenetic_signal_k_with_se(
    tree: ToyTree,
    data: Union[str, Sequence[float]],
    error: Union[str, Sequence[float]] = None,
    test: bool = False,
    permutations: int = 1000,
) -> dict[str, float]:
    """Calculate phylogenetic signal (K) with measurement error.

    This involves fitting a ML model to estimate the rate ...
    """
    # get ntips and the variance-covariance matrix from tree
    ntips = tree.ntips
    V = get_vcv_matrix_from_tree(tree)

    # [optional] get data as features from the tree
    if isinstance(data, str):
        data = tree.get_node_data(data)[:ntips]
    if isinstance(error, str):
        error = tree.get_node_data(error)[:ntips]
    if error is None:
        error = np.repeat(0.0, ntips)

    # validate proper trait format returned as float array
    x = _validate_features(data, max_dim=1, size=ntips)    

    # calculate K stat w/ error
    kstat, sig2, loglik = _calculate_k_with_se(x, V, error)

    # [optional] permutation test statistic
    if permutations:
        pval = _permutation_test_k_with_se(permutations, x, V, error, kstat)

    # return as a dict
    return {
        "K": kstat, 
        "P-value": pval,
        "permutations": 0 if not test else permutations,
        "log-likelihood": -loglik,
        "sig2": sig2,
    }


def _likelihood(theta: float, V: np.ndarray, E: np.ndarray, y: np.ndarray) -> float:
    """Estimate theta by maximizing the likelihood.
    """
    # weight variances by theta and add Error variance
    C = theta * V + E

    # get pgls mean
    IC = np.linalg.inv(C)
    n = len(y)
    a = np.sum(IC @ y) / np.sum(IC)

    # get log determinant of variance covariance matrix
    det = np.linalg.det(C)
    if det <= 0:
        logdet2 = np.log(1e-12)
    else:
        logdet2 = np.log(det) / 2.

    # compute log likelihood
    term = (y - a)
    logL = (
        -term.T @ IC @ term / 2. - n * np.log(2 * np.pi) / 2. - logdet2
    )
    # print(theta, -logL)
    return -logL        




if __name__ == "__main__":

    import toytree
    toytree.set_log_level("DEBUG")

    # generate test data
    tree = toytree.rtree.unittree(ntips=24, seed=123)
    tree.pcm.simulate_continuous_brownian(
        rates=[1.0, 0.1], tips_only=True, seed=123, inplace=True
    )

    # get K
    k = _phylogenetic_signal_k(tree=tree, data="t0", test=True)
    logger.info(k)

    # get K w/ Error
    k = _phylogenetic_signal_k_with_se(tree=tree, data="t0", error="t1", test=True)
    logger.info(k)



#!/usr/bin/env python

"""Add a color_scale to Cartesian axes.

"""

from typing import Optional, Mapping, Any
from loguru import logger
import toyplot
from toytree import ToyTree
from toytree.core import Cartesian
from toytree.core.apis import add_subpackage_method, AnnotationAPI
from toytree.style.src.validate_utils import substyle_dict_to_css_dict
# from toytree.annotate.src.checks import get_last_toytree_mark, assert_tree_matches_mark
from toytree.drawing.src.mark_annotation import AnnotationRect  # AnnotationMarker


logger = logger.bind(name="toytree")


# def add_marker_legend(
#     tree: ToyTree,
#     axes: Cartesian,
#     colormap: toyplot.color.Map,
# ) -> None:
#     """...

#     """
#     # get_last_cartesian_axes_from_toytree()
#     axes = tree._last_axes


def add_color_scale_from_feature(
    tree: ToyTree,
    axes: Cartesian,
    feature: str,
    colormap: toyplot.color.Map = None,
    width: Optional[int] = None,
    angle: float = 0.,
    style: Mapping[str, Any] = None,
) -> AnnotationRect:
    """Return color map...

    Extracts the colormap and data domain range from the ToyTreeMark
    which has a plotted feature.

    Examples
    --------
    >>> c, a, m = tree.draw(node_labels="dist")
    >>> tree.annotate.add_color_scale_from_feature("node_labels", xpos=50, ypos=-50)
    >>> # ...
    """


# @add_subpackage_method(AnnotationAPI)
def add_color_scale(
    axes: Cartesian,
    xpos: float = 50,
    ypos: float = -50,
    colormap: toyplot.color.Map = None,
    min: float = 0.,
    max: float = 1.,
    width: Optional[int] = None,
    angle: float = 0.,
    style: Mapping[str, Any] = None,
) -> AnnotationRect:
    """Return ... with color_scale ...

    Add a color scale to a Canvas as a rectangle filled with a color or
    linear gradient of colors. This is an alternative implementation of
    toyplot's add_color_scale method with the aim of having the color
    scale act as a Mark instead of a Numberline axis.

    Parameters
    ----------
    axes: Cartesian
        ...
    xpos: float
        Horizontal position of color scale bar start.
    ypos: float
        Vertical position of color scale bar start.
    colormap: str, toyplot.color.Map, or None
        ...
    min: float
        Value mapped to the minimum (start) color in map.
    max: float
        Value mapped to the maximum (end) color in map.

    See Also
    --------
    toyplot.coordinates.Cartesian.add_color_scale

    Examples
    --------
    # create a color scale given 
    """
    canvas = axes._scenegraph.source("render", axes)
    canvas.color_scale(
        colormap=colormap,
        x1=self._xmax_range + width + self._padding,
        x2=self._xmax_range + width + self._padding,
        y1=self._ymax_range,
        y2=self._ymin_range,
        width=width,
        padding=padding,
        show=True,
        label=label,
        ticklocator=tick_locator,
        scale="linear",
        )
    return axis    


if __name__ == "__main__":
    pass



#!/usr/bin/env python

"""Call an R script to test a related function in R.
"""

import subprocess
import tempfile
from textwrap import dedent


def run_r_script(r_code: str) -> str:
    """Write an R script to temp, execute it, and return stdout.
    
    Parameters
    ----------    
    r_code: str
        The R code to execute.
    """
    with tempfile.NamedTemporaryFile(mode="w", suffix=".R", delete=False) as tmpfile:
        tmpfile.write(r_code)
        tmpfile_path = tmpfile.name
        tmpfile.flush()

        try:
            cmd = ["Rscript", tmpfile_path]
            result = subprocess.run(cmd, capture_output=True, text=True, check=True)
            return result.stdout
        except subprocess.CalledProcessError as e:
            return f"Error: {e.stderr}"


if __name__ == "__main__":

    # Example usage
    r_script = dedent("""
    x <- rnorm(10)
    print(mean(x))
    """)

    output = run_r_script(r_script)
    print("R Output:", output)



#!/usr/bin/env python

"""Functions for finding/estimating the root of a tree

Methods
-------
- root_on_longest_edge(gene_tree)
- root_on_minimal_ancestor_deviations(...)
- root_on_minimal_duplication_loss_score(species_tree, gene_tree)

References
----------
- Farris, J. S. Estimating phylogenetic trees from distance matrices.
  Am. Nat. 106, 645–668 (1972).
- Tria, F., Landan, G. & Dagan, T. Phylogenetic rooting using
  minimal ancestor deviation. Nat Ecol Evol 1, 0193 (2017).
  https://doi.org/10.1038/s41559-017-0193
- ...
"""

from typing import TypeVar, Dict, Union, Optional, Sequence
import itertools
import numpy as np
from loguru import logger
from toytree import ToyTree, Node
from toytree.core.apis import TreeModAPI, add_subpackage_method
from toytree.utils import ToytreeError

Query = TypeVar("Query", str, int, Node)
logger = logger.bind(name="toytree")


__all__ = [
    "root_on_midpoint",
    "root_on_minimal_ancestor_deviation",
]


@add_subpackage_method(TreeModAPI)
def root_on_midpoint(
    tree: ToyTree,
    inplace: bool = False,
    edge_features: Optional[Sequence[str]] = None,
) -> ToyTree:
    """Return ToyTree rooted on midpoint of longest edge.

    Rooting on the "midpoint" assumes a clock-like evolutionary rate
    (i.e., branch lengths are equal to time) and may yield odd results
    when this assumption is violated. This algorithm calculates the
    pairwise path length between all tips in an unrooted tree and roots
    on the midpoint of the longest path.

    Parameters
    ----------
    tree: ToyTree
        Tree that will be (re-)rooted.
    inplace: bool
        If True then the input tree object is modified and returned,
        else a copy of the tree is modified and returned.
    edge_features: Sequence[str]
        One or more Node features that should be treated as a feature
        of its edge, not the Node itself. On rooting, edge features
        are re-polarized, to apply to the correct Node. The 'dist'
        and 'support' features are always treated as edge features.
        Add additional edge features here. See docs for example.

    References
    ----------
    - Farris, J. S. Estimating phylogenetic trees from distance
    matrices. Am. Nat. 106, 645–668 (1972).

    Example
    -------
    >>> tree = toytree.rtree.unittree(10).unroot()
    >>> rtree = tree.mod.root_by_midpoint()
    """
    # get matrix of pairwise tip distances
    dmat = tree.distance.get_tip_distance_matrix()

    # get any pair of Nodes that span the max distance
    pairs = np.where(dmat == dmat.max())
    n0, n1 = pairs[0][0], pairs[1][0]

    # midpoint is half this distance
    dist_to_new_root = dmat[n0, n1] / 2.

    # going up this dist from one of the two Nodes will hit the
    # pseudo-root, but not for the other. Select the other.
    for idx in [n0, n1]:
        node = tree[idx]
        dist_below = 0.
        while 1:
            # this is the correct edge to root on
            if (node._dist + dist_below) >= dist_to_new_root:
                root_node = node
                root_node_dist = dist_to_new_root - dist_below
                break

            # cannot test any higher, restart from other tip Node.
            if node._up.is_root():
                break

            # add blen and check further up from this Node.
            dist_below += node._dist
            node = node._up

    # return tree or copy re-rooted
    # tree = tree if inplace else tree.copy()
    tree = tree.root(
        root_node.idx,
        root_dist=root_node_dist,
        edge_features=edge_features,
        inplace=inplace,
    )
    return tree


@add_subpackage_method(TreeModAPI)
def root_on_minimal_ancestor_deviation(
    tree: ToyTree,
    *query: Query,
    inplace: bool = False,
    return_stats: bool = False,
    min_dist: float = 1e-12,
    edge_features: Optional[Sequence[str]] = None,
) -> Union[ToyTree, Dict[ToyTree, Dict[str, float]]]:
    """Return ToyTree rooted on edge that minimizes ancestor deviations.

    The minimal ancestor deviation (MAD) rooting method attempts to
    accommodate the effect of rate heterogeneity on branch lengths when
    rooting a tree by branch length information. This is done by using
    pairwise topological and metric information in unrooted trees. It
    assumes that branch lengths are additive and that tips (OTUs) are
    contemporaneous in time (i.e., tip height variation results from
    rate heterogeneity). This method evaluates the deviations of the
    global midpoint root position from each pairwise midpoint rooting
    position and minimizes it.

    Parameters
    ----------
    tree: ToyTree
        A tree with branch lengths on which to infer a MAD rooting.
    query: int, str or Node
        Optional. Select one or more Nodes by int idx label, str name,
        or Node object to specify a branch on which to root. By default
        this is left empty and the optimal rooting is returned. But if
        you wish to obtain a tree rooted on an alternative edge then
        this will still use the MAD algorithm to minimize ancestral
        deviations to find the optimal node position on that edge.
    inplace: bool
        If True return the input tree modified, else return a modified
        copy of the tree.
    return_stats: bool
        If True the returned value is a (ToyTree, Dict) instead of just
        a ToyTree and the Dict contains the 'root_ambiguity_index' and
        the coefficient of variation in MAD. Note: edge features MAD
        and MAD_root_prob are always set to Nodes of the returned tree.
    min_dist: float
        Zero length edges use a min_dist value of default=1e-12 to
        prevent division by zero errors.
    edge_features: Sequence[str]
        One or more Node features that should be treated as a feature
        of its edge, not the Node itself. On rooting, edge features
        are re-polarized, to apply to the correct Node. The 'dist'
        and 'support' features are always treated as edge features.
        Add additional edge features here. See docs for example.

    Note
    ----
    Returns a ToyTree with MAD and MAD_root_prob features mapped to each
    Node as an edge_feature. These represent the minimum ancestor
    deviations if the tree is rooted on each edge, and the relative
    probability that each is the correct root. If rooting is unknown
    then downstream analysis can integrate over alternative root
    hypotheses weighted by their probability.

    Compared to Tria et al's mad.py script this function yields
    identical results when edges are non-zero, but varies *slightly* in
    how zero-length branches affect results.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(4, seed=123).unroot()
    >>> tree.set_node_data('name', dict(zip(range(tree.nnodes), list("abcdCA"))), inplace=True)
    >>> tree.set_node_data('dist', {'b': 1, 'd': 4, 'C': 1}, default=3, inplace=True)
    >>> rtree, stats = root_on_minimal_ancestor_deviation(tree, return_stats=True)
    >>> rtree.draw('p', node_sizes=18, node_labels="MAD")
    >>> print(stats)

    Reference
    ---------
    - Tria, F., Landan, G. & Dagan, T. Phylogenetic rooting using
    minimal ancestor deviation. Nat Ecol Evol 1, 0193 (2017).
    https://doi.org/10.1038/s41559-017-0193
    """
    # tell user to resolve polytomies.
    assert tree.is_bifurcating(include_root=False), (
        "tree contains polytomies, first resolve manually or with "
        "`tree.mod.resolve_polytomies()`.")

    # get a copy of the tree and translatable query
    if query:
        query = tree.get_mrca_node(*query).get_leaf_names()
    tree = tree.unroot(inplace=inplace)

    # warn user about zero len edges
    zero_len_edges = sum(1 for i in tree[:-1] if i._dist < min_dist)
    if zero_len_edges:
        logger.warning(
            f"Using min_dist {min_dist} for {zero_len_edges} zero len edges in tree.")

    # get pairwise dist matrix w/ min_dist for zero len edges
    dmat = tree.distance.get_node_distance_matrix()
    dmat[dmat < min_dist] = min_dist
    dmat[np.diag_indices_from(dmat)] = 0.

    # store paths between pairs of tips, {(0, 2): [1, 5, 2]... }
    paths = {}
    for path in itertools.combinations(range(tree.ntips), 2):
        node_path = tree.distance.get_node_path(*path)
        paths[path] = [i.idx for i in node_path][1:-1]
        paths[path[::-1]] = paths[path][::-1]

    # get npairs and a dict to store the Rbranch statistics
    npairs = int((tree.ntips * (tree.ntips - 1)) / 2)
    rij_dict = {}

    # iterate over (idx, jdx) edges separating (seti, setj) tip sets.
    inodes = tree.iter_edges()
    ibiparts = tree.enum._iter_bipartition_sets(feature="idx", include_singleton_partitions=True)
    for (node_i, node_j), (seti, setj) in zip(inodes, ibiparts):

        # get .idx label of nodes idx,jdx representing the edge
        idx = node_i._idx
        jdx = node_j._idx
        dij = dmat[idx, jdx]

        # iterate over tips descended from the two clades split by the new root
        # and calculate rho as the relative position of root node on the edge.
        rho_top = 0
        rho_bot = 0
        for tipb in seti:
            for tipc in setj:
                dbc = dmat[tipb, tipc]
                dbi = dmat[tipb, idx]
                rho_top += (dbc - (2 * dbi)) * (dbc ** -2)
                rho_bot += (2 * dij) * (dbc ** -2)

        # the relative position of root on edge (idx, jdx) constrained in [0, 1]
        rho = rho_top / rho_bot
        rho_i = min(max(0, rho), 1)

        # distance from i or j to ancestor
        dio = dmat[idx, jdx] * rho_i
        # logger.debug(f"ij={idx, jdx}, dio={dio:.3g}, rho={rho:.3f} rho_i={rho_i} dij={dmat[idx, jdx]}")

        # get sum sq rel deviation for this rooting over all tip pairs
        relative_deviations = np.zeros(npairs)
        pidx = 0

        # iterate over pairs of tips both in seti
        for tipb, tipc in itertools.combinations(seti, 2):
            dbc = dmat[tipb, tipc]
            aidx = max(paths[(tipb, tipc)])
            dab = dmat[tipb, aidx]
            relative_deviations[pidx] = abs(((2 * dab) / dbc) - 1)
            pidx += 1

        # iterate over pairs of tips both in setj above jdx
        above_j = setj - set(i._idx for i in node_j.iter_descendants())
        for tipb, tipc in itertools.combinations(above_j, 2):
            dbc = dmat[tipb, tipc]
            aidx = min(paths[(tipb, tipc)], key=lambda x: dmat[jdx, x])
            dab = dmat[tipb, aidx]
            relative_deviations[pidx] = abs(((2 * dab) / dbc) - 1)
            pidx += 1

        # iterate over pairs of tips both in setj below jdx
        below_j = setj - above_j
        for tipb, tipc in itertools.combinations(below_j, 2):
            dbc = dmat[tipb, tipc]
            aidx = max(paths[(tipb, tipc)])
            dab = dmat[tipb, aidx]
            relative_deviations[pidx] = abs(((2 * dab) / dbc) - 1)
            pidx += 1

        # iterate over pairs of tips both in setj but split above and below jdx
        for tipb, tipc in itertools.product(below_j, above_j):
            dbc = dmat[tipb, tipc]
            aidx = jdx
            dab = dmat[tipb, aidx]
            relative_deviations[pidx] = abs(((2 * dab) / dbc) - 1)
            pidx += 1

        # iterate over pairs spanning the root (idx, jdx) split
        for tipb, tipc in itertools.product(seti, setj):
            dbc = dmat[tipb, tipc]
            aidx = idx
            dab = dmat[tipb, aidx] + dio
            relative_deviations[pidx] = abs(((2 * dab) / dbc) - 1)
            pidx += 1

        # make sum-squares into root-mean-square and save to rij_dict
        rbranch = np.sqrt(np.mean(relative_deviations ** 2))

        # warn user if rbranch exceeds 1.
        if rbranch > 1:
            constrained_rbranch = min(max(0, rbranch), 1)
            logger.warning(f"edge {idx, jdx} MAD={rbranch:.2f} outlier constrained to {constrained_rbranch:.0f}))")
            rbranch = constrained_rbranch
        # logger.info(f"edge {idx, jdx} rho={rho_i:.3f} R={rbranch:.4f}")
        rij_dict[(idx, jdx, rho_i)] = rbranch

    # root ambiguity index is ratio of best to second best
    r_sorted = sorted(rij_dict, key=lambda x: rij_dict[x])
    root_ambiguity_index = rij_dict[r_sorted[0]] / rij_dict[r_sorted[1]]

    # warn user if equally likely rootings
    if root_ambiguity_index == 1:
        logger.warning(
            ">1 optimal rootings exist. Check the 'MAD' and 'MAD_root_prob' "
            "features on the returned tree to examine alternative rootings.")

    # select node to root the tree on. Choose the optimal MAD edge.
    if not query:
        ridx, uidx, rho_i = r_sorted[0]
        logger.info(f"rooting on MAD edge {ridx, uidx} w/ optimized rho={rho_i:.2f}")

    # or, root on user-selected edge which may not be the optimal
    else:
        root_node = tree.get_mrca_node(*query)
        if root_node.is_root():
            raise ToytreeError(f"Cannot root on current root {root_node}, select another Node.")
        ridx = root_node._idx

        # find the rho value for this edge rooting
        for edge in r_sorted:
            if edge[0] == ridx:
                _, _, rho_i = edge
        logger.info(f"rooting on user-selected edge {ridx, root_node._up._idx} w/ optimized rho={rho_i:.2f}")

    # set values to Nodes
    root_probs = {}
    mads = {}
    summed = sum(1 - i for i in rij_dict.values())
    for edge, value in rij_dict.items():
        nidx, _, _ = edge
        mads[nidx] = value
        root_probs[nidx] = (1 - value) / summed
    tree.set_node_data("MAD", mads, inplace=True)
    tree.set_node_data("MAD_root_prob", root_probs, inplace=True)

    # add these features as edge features
    tree.edge_features = tree.edge_features | {"MAD", "MAD_root_prob"}

    # root the tree
    tree.root(
        ridx,
        root_dist=tree[ridx]._dist * rho_i,
        edge_features=edge_features,
        inplace=True,
    )

    # clock-likeness of inferred root position is described by the root
    # clock coefficient of variation on the new rooted tree. Note the
    # usage of ddof=1 to match the results of Tria et al.
    dmat = tree.distance.get_node_distance_matrix()
    dists = np.array([dmat[i, tree.treenode._idx] for i in range(tree.ntips)])
    root_clock_coefficient_of_variation = 100 * (dists.std(ddof=1) / dists.mean())
    stats = dict(
        minimal_ancestor_deviation=rij_dict[r_sorted[0]],
        root_ambiguity_index=root_ambiguity_index,
        root_clock_coefficient_of_variation=root_clock_coefficient_of_variation,
    )
    logger.info(f"stats={stats}")
    if return_stats:
        return tree, stats
    return tree


if __name__ == "__main__":

    import toytree
    toytree.set_log_level("INFO")
    import numpy as np

    # tree = toytree.rtree.unittree(6, seed=123).unroot()
    # tree = toytree.rtree.rtree(5, seed=123).mod.ladderize()#.unroot()
    # tree.set_node_data('dist', {4: 5, 7: 3}, inplace=True)
    # tree.set_node_data('name', {5: "X", 6: "Y", 7: "Z"}, inplace=True)
    # tree.treenode.draw_ascii()
    # tree.write("/tmp/test.tree")

    # test1 rtree 5-tips variable edgelens
    # tree = toytree.rtree.rtree(5, seed=123).unroot()
    # tree.set_node_data('name', {i: j for (i, j) in enumerate("abcdeXYR")}, inplace=True)
    # tree.set_node_data('dist', {'e': 5, 'Y': 3}, inplace=True)
    # tree.write("/tmp/test1.tree")
    # c1, a, m = tree.draw(ts='p', layout='r', node_hover=True)
    # rtree, stats = root_on_minimal_ancestor_deviation(tree, return_stats=True)
    # c2, a, m = rtree.draw(ts='p', layout='r', node_hover=True)
    # print(stats)
    # print(rtree.get_node_data())
    # toytree.utils.show([c1, c2])

    # test1 alt-rooting rtree 5-tips variable edgelens
    # tree = toytree.rtree.rtree(5, seed=123).unroot()
    # tree.set_node_data('name', {i: j for (i, j) in enumerate("abcdeXYR")}, inplace=True)
    # tree.set_node_data('dist', {'e': 5, 'Y': 3}, inplace=True)
    # tree.write("/tmp/test1.tree")
    # c1, a, m = tree.draw(ts='p', layout='r', node_hover=True)
    # rtree, stats = root_on_minimal_ancestor_deviation(tree, 'a', return_stats=True)
    # c2, a, m = rtree.draw(ts='p', layout='r', node_hover=True)
    # print(stats)
    # print(rtree.get_node_data())
    # toytree.utils.show([c1, c2])

    # test2
    # tree = toytree.rtree.rtree(5, seed=123).unroot()
    # tree.set_node_data('name', {i: j for (i, j) in enumerate("abcdeXYR")}, inplace=True)
    # tree.set_node_data('dist', {'e': 5, 'Y': 3, 'X': 10}, inplace=True)
    # tree.write("/tmp/test2.tree")
    # c1, a, m = tree.draw(ts='p', layout='r', node_hover=True)
    # rtree, stats = root_on_minimal_ancestor_deviation(tree, return_stats=True)
    # c2, a, m = rtree.draw(ts='p', layout='r', node_hover=True)
    # print(stats)
    # print(rtree.get_node_data())
    # toytree.utils.show([c1, c2])

    # test w/ polytomies
    tree = toytree.rtree.rtree(10, seed=123).unroot()
    tree.mod.collapse_nodes(1, 2, 12, 15, inplace=True)
    tree.set_node_data('dist', {2: 5, 12: 3, 13: 10}, inplace=True)
    tree.write("/tmp/test3.tree")
    c1, a, m = tree.draw(ts='p', layout='r', node_hover=True)

    tree = tree.mod.resolve_polytomies()
    rtree, stats = root_on_minimal_ancestor_deviation(tree, return_stats=True)
    c2, a, m = rtree.draw(ts='p', layout='r', node_hover=True)
    print(stats)
    print(rtree.get_node_data())
    toytree.utils.show([c1, c2])



#!/usr/bin/env python

"""...

"""

from toytree.layout.src.layout_tip_angles import get_tip_labels_angles
from toytree.layout.src.layout_base import BaseLayout
from toytree.layout.src.layout_linear import LinearLayout
from toytree.layout.src.layout_circular import CircularLayout
from toytree.layout.src.layout_unrooted import UnrootedLayout



#!/usr/bin/env python

"""Infer a consensus tree from a set of trees.

This can be performed with a set of trees that all share the same tips.
(See supertree methods for trees with non-overlapping tip sets.)

Examples
--------
# get consensus tree with supports.
>>> trees = ...
>>> ctree = toytree.infer.get_majority_rule_consensus_tree(trees)
>>> print(ctree.get_node_data())

# get consensus tree with supports and conditional edge dist values
>>> trees = ...
>>> ctree = toytree.infer.get_majority_rule_consensus_tree(trees)
>>> cctree = toytree.infer.map_features_to_consensus_tree(ctree, trees, conditional=True)
>>> print(cctree.get_node_data())

# get consensus tree of rooted trees with support and conditional node heights
>>> trees = ...
>>> ctree = toytree.infer.get_majority_rule_consensus_tree(trees)
>>> rtree = ctree.root(trees[0].treenode.children[0].get_leaf_names())
>>> ftree = toytree.infer.map_features_to_consensus_tree(rtree, trees, rooted=True)
>>> print(ftree.get_node_data())


TODO
----
- support mapping arbitrary traits (discrete and quantitative?) from
  a set of trees onto the nodes of a consensus tree.

"""

from typing import TypeVar, Union, List, Dict
import numpy as np
from toytree.core import ToyTree, Node
from loguru import logger

MultiTree = TypeVar("MultiTree")
logger = logger.bind(name="toytree")

__all__ = [
    "get_consensus_tree",     # (trees)
    "get_consensus_features"  # (tree, trees, ...)
]


def check_trees_set_for_ultrametric(trees: List[ToyTree]) -> None:
    """Raise exception if input trees are not suitable for consensus
    tree analysis with option ultrametric=True.
    """
    for tre in trees:
        if not tre.is_rooted():
            msg = "input trees are not rooted. Cannot use option 'ultrametric=True'"
            logger.error(msg)
            raise ValueError(msg)
        if not np.allclose([i._height for i in tre[:tre.ntips]], 0.0, atol=1e-5):
            msg = "input trees are not ultrametric. Cannot use option 'ultrametric=True'"
            logger.error(msg)
            raise ValueError(msg)


def get_clade_frequencies(trees: Union[MultiTree, List[ToyTree]]) -> Dict[frozenset, float]:
    """Return a dict mapping bipartitions to their frequencies and dists.

    This performs one pass through each tree to get its bipartitions
    and records the frequency of each split and the dist of edges
    corresponding to the split. Rooted trees root node position is
    ignored.

    Parameters
    ----------
    trees: MultiTree | list[ToyTree]
        A collection of ToyTrees (list or MultiTree object) sharing the
        same tip labels.
    """
    # require all trees to share the same tips
    # ... TODO: use MultiTree... or do this outside this func.

    # dict to store clade occurrences. Set full tip set to freq=1.0.
    ntrees = len(trees)
    all_tips = frozenset(trees[0].get_tip_labels())
    clades = {all_tips: {"count": ntrees}}

    # iterate over the input tree set
    for tre in trees:

        # iterate over splits in the tree as (fset, fset)
        iter_biparts = tre.iter_bipartitions("name", True, False, type=frozenset, sort=True)
        for node, bipart in zip(tre, iter_biparts):

            # get node dist; extra if root edge and tree is rooted.
            dist = node._dist
            if not node._up.is_root():
                if tre.is_rooted():
                    dist = sum(node._dist for i in node._up.children)

            # store the partition count and dist
            part = bipart[0]
            if part not in clades:
                clades[part] = {"count": 1, "dist": [dist]}
            else:
                clades[part]["count"] += 1
                clades[part]["dist"].append(dist)

    # sort clades by occurrence
    sclades = sorted(clades, key=lambda x: clades[x]["count"], reverse=True)

    # sort and convert counts to (frequency of trees containing the split)
    clades = {i: clades[i] for i in sclades}
    for clade in clades:
        clades[clade]["freq"] = clades[clade]["count"] / ntrees
    return clades


def get_consensus_clades(clades: Dict[frozenset, float], min_freq: float) -> Dict[frozenset, float]:
    """Return dict with only non-conflicting clades above min_freq.

    This is used prior to constructing a majority-rule consensus tree,
    and min_freq sets the minimum clade frequency to keep/collapse.
    """
    keep = {}
    mark = []
    for clade, data in clades.items():
        freq = data["freq"]
        if freq < min_freq:
            continue

        # bipart conflicts with others, discard.
        conflict = False
        for c in keep:
            if conflict:
                break
            if c.isdisjoint(clade):
                continue
            if c.issubset(clade):
                continue
            if c.issuperset(clade):
                continue
            conflict = True

        # keep this bipartition
        if not conflict:
            keep[clade] = data
        # if conflicted w/ equal frequency then mark existing for removal
        elif conflict and (freq == keep[c]["freq"]):
            mark.append(c)

    # remove any clades that conflicted w/ equal frequency
    for clade in mark:
        if clade in keep:
            keep.pop(clade)
    return keep


def build_consensus_tree(clades: Dict[frozenset, float]) -> ToyTree:
    """Return a majority-rule consensus tree constructed from non-
    conflicting clades.
    """
    # sort clades by size, large to small
    sclades = sorted(clades, key=len, reverse=True)

    # dict {clade-set: Node} in order they are added. Root Node first.
    treenode = sclades.pop(0)
    sets_to_nodes = {treenode: Node()}

    # iterate over filtered clades ordered by size
    for clade in sclades:

        # create the Node
        name = str(*clade) if len(clade) == 1 else ""
        node = Node(name=name, support=clades[clade]["freq"], dist=np.mean(clades[clade]["dist"]))

        # connect node to smallest clade parent
        # TODO: probably faster not to have to sort here
        for eclade in sorted(sets_to_nodes, key=len):
            if clade.issubset(eclade):
                enode = sets_to_nodes[eclade]
                enode._add_child(node)
                break
        # store this {clade: node}
        sets_to_nodes[clade] = node

    # convert to ToyTree
    tree = ToyTree(sets_to_nodes[treenode])

    # set tip and root supports to nan
    tree[-1].support = np.nan
    for nidx in range(tree.ntips):
        tree[nidx].support = np.nan
    return tree


# def get_clade_frequencies2(trees: MultiTree, rooted: bool) -> dict[tuple, float]:
#     """Return a dict mapping bipartitions to their frequencies and dists.

#     This performs one pass through each tree to get its bipartitions
#     and records the count of each split and stores the dist or height
#     associated with the clade. If trees are rooted and ultrametric then
#     heights are stored.

#     Parameters
#     ----------
#     trees: MultiTree
#         A MultiTree containing multiple ToyTree objects.
#     rooted: bool
#         If True node heights are stored instead of dists.
#     """
#     clades = {}
#     ntrees = len(trees)

#     def feat(node: Node) -> float:
#         if rooted:
#             return getattr(node, "_height")
#         return getattr(node, "_dist")

#     # iterate over splits in input trees. If the tree is rooted then
#     # store clades of both descendants from root, not just the split.
#     for tre in trees:
#         iter_biparts = tre.iter_bipartitions("name", True, False, type=frozenset, sort=True)
#         for node, bipart in zip(tre, iter_biparts):
#             part = bipart[0]
#             if part not in clades:
#                 clades[part] = {feature: [feat(node)], "count": 1}
#             else:
#                 clades[part]["count"] += 1
#                 clades[part][feature].append(feat(node))

#             # for rooted trees store other root child
#             if rooted and node.up.is_root():
#                 node = node.get_sisters()[0]
#                 part = bipart[1]
#                 if part not in clades:
#                     clades[part] = {feature: [feat(node)], "count": 1}
#                 else:
#                     clades[part]["count"] += 1
#                     clades[part][feature].append(feat(node))

#         # add full clade
#         part = frozenset(tre.get_tip_labels())
#         if part not in clades:
#             clades[part] = {feature: [feat(tre.treenode)], "count": 1}
#         else:
#             clades[part]["count"] += 1
#             clades[part][feature].append(feat(tre.treenode))

#     # sort clades by occurrence
#     sclades = sorted(clades, key=lambda x: clades[x]["count"], reverse=True)

#     # sort and convert counts to (frequency of trees containing the split)
#     clades = {i: clades[i] for i in sclades}
#     for clade in clades:
#         clades[clade]["freq"] = clades[clade]["count"] / ntrees
#     clades[part]["freq"] = np.nan
#     return clades

# def build_consensus_tree2(clades: dict[frozenset, float], rooted: bool) -> ToyTree:
#     """Return a majority-rule consensus tree constructed from non-
#     conflicting clades.
#     """
#     # dict mapping {clade-set: Node} in order they are added.
#     sets_to_nodes = {}

#     # sort clades by size
#     sclades = sorted(clades, key=len, reverse=True)
#     for clade in sclades:
#         if clade in sets_to_nodes:
#             continue

#         # create the Node
#         name = str(*clade) if len(clade) == 1 else ""
#         node = Node(name=name, support=clades[clade]["freq"])
#         # feats[node.idxname] = clades[clade][feature]
#         node.feat = clades[clade][feature]

#         # connect nodes (visit existing smallest to largest)
#         for eclade in sorted(sets_to_nodes, key=len):
#             if clade.issubset(eclade):
#                 enode = sets_to_nodes[eclade]
#                 enode._add_child(node)
#                 break
#         # store this {clade: node}
#         sets_to_nodes[clade] = node

#     # convert to ToyTree
#     tree = ToyTree(sets_to_nodes[sclades[0]])

#     # set node dist and height attrs
#     if rooted:
#         tree.set_node_data("height", {i: np.mean(i.feat) for i in tree}, inplace=True)
#         tree.set_node_data("height_min", {i: np.min(i.feat) for i in tree}, inplace=True)
#         tree.set_node_data("height_max", {i: np.max(i.feat) for i in tree}, inplace=True)
#         tree.set_node_data("height_std", {i: np.std(i.feat) for i in tree}, inplace=True)
#     else:
#         tree.set_node_data("dist", {i: np.mean(i.feat) for i in tree}, inplace=True)
#         tree.set_node_data("dist_min", {i: np.min(i.feat) for i in tree}, inplace=True)
#         tree.set_node_data("dist_max", {i: np.max(i.feat) for i in tree}, inplace=True)
#         tree.set_node_data("dist_std", {i: np.std(i.feat) for i in tree}, inplace=True)
#     tree.remove_features(feature, inplace=True)
#     return tree


def get_consensus_features(
    tree: ToyTree,
    trees: MultiTree,
    features: List[str] = None,
    ultrametric: bool = False,
    conditional: bool = False,
) -> ToyTree:
    """Return tree with feature data mapped to each bipartition from
    a set of trees that may or may not share the same bipartitions.

    The 'support' value on the consensus tree represents the proportion
    of trees that supported the split in the consensus, and the 'dist'
    is the mean dist value of the bipartition edge in the set of trees
    that included that edge. These two features are set on MJ rule
    consensus trees by default. Other features can be optionally
    computed as well.

    Parameters
    ----------
    tree: ToyTree
        The tree can be the MJ consensus tree or a user input tree.
    trees: MultiTree | list[ToyTree]
        A list of trees from which features will be extracted. Support
        is always measured. Edge lengths are extracted as 'dists' or
        'heights' depending on the option 'rooted'.
    features: None | list[str]
        A list of feature names that exist on the input trees that you
        wish to have summarized across the nodes of the consensus tree.
        For quantitative features this will record min, max, mean, std
        conditional on the existence of the node in the tree.
    ultrametric: bool
        If trees are rooted and ultrametric then set this option to
        True to summarize stats of node heights instead of node dists.
        Note: this forces conditional=True.
    conditional: bool
        If True then dist values of tip nodes on unrooted trees are
        only calculated from the subset of trees where these tips occur
        in the same splits as in the main tree, otherwise dists are
        calculated from tip nodes across all input trees. This only
        affects tip node dists.
    """
    if ultrametric:
        # mtree.all_tree_tips_aligned()
        return map_rooted_tree_supports_and_heights_to_rooted_tree(tree, trees, features)
    return map_unrooted_tree_supports_and_dists_to_unrooted_tree(tree, trees, features, conditional)


def map_unrooted_tree_supports_and_dists_to_unrooted_tree(
    tree: ToyTree,
    trees: MultiTree,
    features: List[str] = None,
    conditional: bool = False,
) -> ToyTree:
    """
    """
    # save outgroup to re-root tree later
    outg = tree.treenode.children[1].get_leaf_names()

    # create a copy of the input tree and set support values to zero
    tree = tree.unroot().set_node_data("support", default=np.nan)

    # get the non-singleton bipartitions
    biparts = list(tree.iter_bipartitions("name", False, False, type=frozenset, sort=True))

    # iterate over all input trees
    data = {}
    tips = {}
    for tre in trees:
        
        # get this trees non-singleton biparts
        biparts2 = tre.iter_bipartitions("name", False, False, type=frozenset, sort=True)
        
        # iterate over biparts in test tree
        for node, bipart in zip(tre[tre.ntips:], biparts2):
            
            # record the bipartition counts to get supports and dist | height
            if bipart in biparts:
                if bipart not in data:
                    data[bipart] = {"count": 1, "dist": [node._dist]}
                else:
                    data[bipart]["count"] += 1
                    data[bipart]["dist"].append(node._dist)

            # if this node contains tips as children then store their
            # features as well conditional on the existence of this clade.
            if conditional:
                for child in node._children:
                    if child.is_leaf():
                        if child.name not in tips:
                            tips[child.name] = {"count": 1, "dist": [node._dist]}
                        else:
                            tips[child.name]["count"] += 1
                            tips[child.name]["dist"].append(node._dist)

                # if tree is unrooted and node parent is root then its
                # sisters can be tips, supported by this split.
                if node.up.is_root():
                    for child in node.get_sisters():
                        if child.is_leaf():
                            if child.name not in tips:
                                tips[child.name] = {"count": 1, "dist": [node._dist]}
                            else:
                                tips[child.name]["count"] += 1
                                tips[child.name]["dist"].append(node._dist)

        # if not conditional tip dists then add all children here.
        if not conditional:
            for node in tre[:tre.ntips]:
                if node.name not in tips:
                    tips[node.name] = {"dist": [node._dist]}
                else:
                    tips[node.name]["dist"].append(node._dist)

    # iterate over input tree nodes and splits
    for node, bipart in zip(tree[tree.ntips:], biparts):
        if bipart in data:
            node.support = data[bipart]["count"] / len(trees)
            setattr(node, "dist_mean", np.mean(data[bipart]["dist"]))
            setattr(node, "dist_min", np.min(data[bipart]["dist"]))
            setattr(node, "dist_max", np.max(data[bipart]["dist"]))
            setattr(node, "dist_std", np.std(data[bipart]["dist"]))
            node._dist = node.dist_mean
            # node.name = set(node.get_leaf_names())            

    for name in tips:
        node = tree.get_nodes(name)[0]
        setattr(node, "dist_mean", np.mean(tips[node.name]["dist"]))
        setattr(node, "dist_min", np.min(tips[node.name]["dist"]))
        setattr(node, "dist_max", np.max(tips[node.name]["dist"]))
        setattr(node, "dist_std", np.std(tips[node.name]["dist"]))
        node._dist = node.dist_mean

    # set as edge features in case user re-roots the tree
    tree.edge_features.add("dist_mean")
    tree.edge_features.add("dist_min")
    tree.edge_features.add("dist_max")
    tree.edge_features.add("dist_std")
    # tree = tree.mod.root_on_minimal_ancestor_deviation(*outg)
    return tree


def map_rooted_tree_supports_and_heights_to_rooted_tree(
    tree: ToyTree,
    trees: MultiTree,
    features: List[str] = None,
    conditional: bool = False,
) -> ToyTree:
    """

    """
    # check that rooted is valid
    check_trees_set_for_ultrametric(trees)

    # create a copy of the input tree and set support values to zero
    tree = tree.copy().set_node_data("support", default=np.nan)

    # store clades as {node: frozenset(leaf_names), ...} for internal nodes
    clades = [frozenset(i.get_leaf_names()) for i in tree[tree.ntips:]]

    # iterate over all input trees
    data = {}
    # tips = {}
    for tre in trees:

        # get this trees clades
        tclades = {i: frozenset(i.get_leaf_names()) for i in tre[tre.ntips:]}

        # iterate over nodes in rooted trees
        for node, clade in tclades.items():
            if clade not in data:
                data[clade] = {"count": 1, "height": [node._height]}
            else:
                data[clade]["count"] += 1
                data[clade]["height"].append(node._height)

    # iterate over input tree nodes and splits
    for node, clade in zip(tree[tree.ntips:], clades):
        if clade in data:
            node.support = data[clade]["count"] / len(trees)
            setattr(node, "height_mean", np.mean(data[clade]["height"]))
            setattr(node, "height_min", np.min(data[clade]["height"]))
            setattr(node, "height_max", np.max(data[clade]["height"]))
            setattr(node, "height_std", np.std(data[clade]["height"]))
    tree.set_node_data("height", {i: i.height_mean for i in tree[tree.ntips:]}, inplace=True)
    tree[-1].support = np.nan
    tree[-2].support = np.nan
    return tree


def get_consensus_tree(trees: Union[MultiTree, List[ToyTree]], min_freq: float=0.0) -> ToyTree:
    """Return an exteded majority-rule consensus tree from a list of trees.

    The trees must contain the same set of tips. The returned tree will
    contain the most frequently occurring non-conflicting clades in the
    input trees. If the most frequent clades conflict with equal
    frequency the node is collapsed. A minimum clade frequency can
    be set to only include clades occurring above that threshold. For
    example, min_freq=0.5 will procude a 50% majority-rule consensus
    tree. The returned tree is unrooted with edge dist values as the
    mean dist across all occurrences of that split in the set of trees.

    Parameters
    ----------
    trees: MultiTree | list[ToyTree]
        A MultiTree or list of ToyTrees sharing the same tip labels.
    min_freq: float
        A minimum frequency cutoff for a split to occur across the set
        of trees for it to be included in the consensus tree.

    See Also
    --------
    map_features_to_consensus_tree
        Map dists, heights, or other features from a set of input trees
        to nodes of a consensus tree with additional options.
    """
    # infer the majority-rule consensus unrooted tree w/ support and dist
    clade_freqs = get_clade_frequencies(trees)
    fclade_freqs = get_consensus_clades(clade_freqs, min_freq)
    ctree = build_consensus_tree(fclade_freqs)
    return ctree


if __name__ == "__main__":

    import toytree
    import numpy as np

    # multiple input rooted trees
    rtrees = toytree.mtree([
        "((a:1,b:1):3,(c:3,(d:2,e:2):1):1);",
        "((a:1,b:1):3,(c:3,(d:2,e:2):1):1);",
        # "((a:1,b:1):4,(e:2,(c:3,d:2):1));",
        # "((a:1,b:1):4,(d:2,(c:3,e:2):1));",
        "((a:1,b:1):3,(e:3,(c:2,d:2):1):1);",
        "((a:1,b:1):4,(d:3,(c:2,e:2):1):2);",
    ])

    # multiple input unrooted trees
    utrees = toytree.mtree([
        "((a:1,b:1):3,(c:3,(d:2,e:2):1):1);",
        "((a:1,b:1):3,(c:3,(d:2,e:2):1):1);",
        # "((a:1,b:1):4,(e:2,(c:3,d:2):1));",
        # "((a:1,b:1):4,(d:2,(c:3,e:2):1));",
        "((a:1,b:1):3,(e:2,(c:3,d:2):1):1);",
        # "((a:1,b:1):3,(e:2,(c:3,d:2):1):1);",        
        "((a:1,b:1):3,(d:2,(c:3,e:2):1):1);",
        # "((d:2,(c:3,e:2):1),(a:1,b:1):3):1;",
        # "((a:1,e:1):3,(d:3,(c:2,b:2):1):1);",
    ])
    utrees = [i.unroot() for i in utrees]

    utrees[0]._draw_browser(node_sizes=16, node_markers="r1.5x1", node_labels="support", tmpdir="~", scale_bar=True)
    # raise SystemExit(0)

    TREES = utrees
    ROOTED = 0

    cdict = get_clade_frequencies(TREES)
    for i in cdict:
        print(i, cdict[i])
    print("-------------")

    cdict = get_consensus_clades(cdict, 0.0)
    for i in cdict:
        print(i, cdict[i])
    print("-------------")

    tree = build_consensus_tree(cdict)
    print(tree.get_node_data())
    tree.treenode.draw_ascii()

    ctree = get_consensus_features(tree, TREES)
    print(tree.get_node_data())


    fish = toytree.mtree("https://eaton-lab.org/data/densitree.nex")
    fish = get_consensus_tree(fish)
    print(fish.get_node_data())
    raise SystemExit(0)    
    # tree = tree.root("a", "b")
    # tree = tree.mod.root_on_minimal_ancestor_deviation()
    # tree[-2].support = np.nan
    # tree._draw_browser(node_sizes=16, node_markers="r1.5x1", node_labels="support", tmpdir="~", scale_bar=True)

    # tree = map_features_to_consensus_tree(tree, trees, None, False, True)
    # print(tree.get_node_data())
    # tree._draw_browser(node_sizes=16, node_markers="r1.5x1", node_labels="support", tmpdir="~", scale_bar=True)

    # map support and dist to correct utree
    tree = map_features_to_consensus_tree(utrees[0], utrees, None, False, False)
    print(tree.get_node_data())
    tree.treenode.draw_ascii()

    # map support and dist to incorrect utree
    tree = map_features_to_consensus_tree(utrees[-1], utrees, None, False, True)
    print(tree.get_node_data())
    tree.treenode.draw_ascii()

    # map support and height to correct rtree
    # tree = map_features_to_consensus_tree(trees[0], trees, None, True, False)
    tree = map_rooted_tree_supports_and_heights_to_rooted_tree(rtrees[0], rtrees, )
    print(tree.get_node_data(["idx", "name", "support", "dist", "height", "height_mean", "height_std"]))
    tree.treenode.draw_ascii()
    # tree._draw_browser(node_sizes=16, node_markers="r1.5x1", node_labels="support", tmpdir="~", scale_bar=True, ts='b')

    # map support and height to incorrect rtree
    tree = map_rooted_tree_supports_and_heights_to_rooted_tree(rtrees[-1], rtrees, )
    print(tree.get_node_data(["idx", "name", "support", "dist", "height", "height_mean", "height_std"]))
    tree.treenode.draw_ascii()
    # tree._draw_browser(node_sizes=16, node_markers="r1.5x1", node_labels="support", tmpdir="~", scale_bar=True, ts='b')

    # map support of unrooted trees onto a rooted tree.
    # should this unroot the tree?
    # should it mess with dist values?
    tree = map_unrooted_tree_supports_and_dists_to_unrooted_tree(rtrees[0], utrees)
    print(tree.get_node_data())
    tree.treenode.draw_ascii()
    tree._draw_browser(node_sizes=16, node_markers="r1.5x1", node_labels="support", tmpdir="~", scale_bar=True, ts='b')



#!/usr/bin/env python

"""Return midpoint coordinate of edges given the layout.

"""

import numpy as np


def get_edge_midpoints(
    etable: np.ndarray,
    ntable: np.ndarray,
    layout: str,
    edge_type: str,
) -> np.ndarray:
    """Return mid coords (x,y) on edges for the add_edge_[x] functions.

    Finding midpoints requires information about the layout and edge_type
    """
    nnodes = ntable.shape[0]
    midpoints = np.zeros((nnodes, 2))
    for ridx in range(etable.shape[0]):
        cidx, pidx = etable[ridx]
        cx, cy = ntable[cidx]
        px, py = ntable[pidx]

        # unrooted layout is always edge_type='c'
        if edge_type == "c":
            midx = (px + cx) / 2.
            midy = (py + cy) / 2.
        else:
            if layout in ("u", "d"):
                midx = cx
                midy = (py + cy) / 2.
            elif layout in ("r", "l"):
                midx = (cx + px) / 2.
                midy = cy
            elif layout == "c":
                raise NotImplementedError("TODO. For now, use w/ edge_type='c'.")
            else:  # "unrooted":
                raise NotImplementedError("TODO. For now, use w/ edge_type='c'.")
        midpoints[cidx] = (midx, midy)
    return midpoints


if __name__ == "__main__":

    import toytree
    tree = toytree.rtree.unittree(10, seed=123)
    c, a, m = tree.draw()
    coords = get_edge_midpoints(m.etable, m.ntable, m.layout, m.edge_type)
    print(coords)



#!/usr/bin/env python

"""Utilities for working with colors.

"""

from typing import Optional, TypeVar
import itertools
import toyplot.color

Palette = TypeVar("Palette")

# default toytree color palettes
COLORS1 = toyplot.color.brewer.palette("Set2")
COLORS2 = toyplot.color.brewer.palette("Dark2")


def color_cycler(palette: Optional[Palette] = None):
    """Return an infinite cycling iterator over colors in a palette.

    This returns a generator that will cycle the palette infinitely. 
    If palette is None it uses `toytree.color.COLORS1` palette.

    Note
    ----
    Do NOT call `list(color_cycler())` since it is infinite. Instead
    you should iterate from the generator with `next`.

    Example
    --------
    >>> tree = toytree.rtree.unittree(25)
    >>> icolors = toytree.color.color_cycler(toytree.color.COLORS2)
    >>> tipcolors = [next(icolors) for i in tree.get_tip_labels()]
    >>> tree.draw(tip_labels_colors=tipcolors)
    """
    if palette is None:
        palette = COLORS1
    return itertools.cycle(palette)


# def split_rgba_to_rgb(style):
#     """TODO: update/simplify using ToyColor..."""
#     for key in style:
#         if key in ["fill", "stroke"]:
#             value = style.pop(key)
#             # value = ColorKit(value)


# def split_rgba_style(style: Dict[str, str]) -> Dict[str, str]:
#     """Split rgba to rgb and opacity.

#     Because many applications (Inkscape, Adobe Illustrator, Qt) don't handle
#     CSS rgba() colors correctly this function does a work-around.
#     Takes a CSS color in rgba, e.g., 'rgba(40.0%,76.1%,64.7%,1.000)'
#     labeled in a dictionary as {'fill': x, 'fill-opacity': y} and
#     returns with fill as rgb and fill-opacity from rgba. Similarly
#     applied to stroke, stroke-opacity.

#     Note
#     -----
#     If the user sets fill-opacity or stroke-opacity in a style dict
#     it overrules/clobbers any value in the 'a' of the rgba. 
#     """
#     if "fill" in style:
#         color = style["fill"]

#         # try to convert color to css with toyplot. Fails on ...
#         try:
#             color = toyplot.color.css(color)
#         except (TypeError, AttributeError):
#             # print(type(color), color)
#             pass

#         # set 'fill' and 'fill-opacity'
#         if str(color) == "none":
#             style["fill"] = "none"
#             style["fill-opacity"] = 1.0
#         else:
#             rgb = "rgb({:.3g}%,{:.3g}%,{:.3g}%)".format(
#                 color["r"] * 100,
#                 color["g"] * 100,
#                 color["b"] * 100,
#             )
#             style["fill"] = rgb
#             style["fill-opacity"] = str(color["a"])

#     if "stroke" in style:
#         color = style["stroke"]
#         try:
#             color = toyplot.color.css(color)
#         except (TypeError, AttributeError):
#             # print(type(color), color)
#             pass

#         if str(color) == "none":
#             style["stroke"] = "none"
#             style["stroke-opacity"] = 1.0
#         else:
#             rgb = "rgb({:.3g}%,{:.3g}%,{:.3g}%)".format(
#                 color["r"] * 100,
#                 color["g"] * 100,
#                 color["b"] * 100,
#             )
#             style["stroke"] = rgb
#             style["stroke-opacity"] = str(color["a"])
#     return style


# def concat_style_to_str(style: Dict[str, str]) -> str:
#     """Return a style dict concatenated into a style string.

#     Example
#     -------
#     >>> x = {'fill': 'rgb(100%,0%,0%)', 'fill-opacity': 1.0}
#     >>> print(style_to_string(x))
#     >>> # 'fill:rgb(100%,0%,0%);fill-opacity:1.0'
#     """
#     if not style:
#         return ""
#     strs = [
#         f"{key}:{value}" for key, value in sorted(style.items())
#         if value is not None
#     ]
#     return ";".join(strs)


if __name__ == "__main__":

    col = toyplot.color.rgba(1.0, 0.5, 0.5, 0.5)
    sty = {"fill": col, "width": "20px", "height": "20px"}
    print(sty)
    # ssty = split_rgba_style(sty)
    # print(ssty)
    # print(style_to_string(ssty))



#!/usr/bin/env python

"""The utils submodule contains a collection of utility tools.

The functions and classes in `utils` have not yet been developed into
a full submodule in `toytree`, but may be moved eventually. Current
utilities exist for the following topics:

Methods
-------
toytree_sequence(Union[tree_sequence, str])
    Return a ToyTreeSequence object for extracting and drawing trees
    from a tree sequence object produced by msprime or SliM3.
parse_network(str)
    Return a ToyTree and admixture dictionary representing the major
    topology (highest gamma branches) and admixture edges (src, dest
    tuples representing lower gamma branches) from a network file
    produced by SNaQ.
normalize_values(np.ndarray)
    Return an array of values that have been binned into fewer
    discrete integer values. This is useful for plotting edge widths
    (e.g., values look nice in range 2-10) from data in some other
    range (e.g., Ne values in range 1e5 to 1e7).


Notes
------
Organization of code in the utils submodule.

utils/src/
    A folder with the utils source code organized into subsubmodules. 
    The functions intended for use by users are accessible from the 
    top level `toytree.utils`. 
utils/src/tree_sequences/
    The ToyTreeSequence class for working with and visualizing 
    tree_sequence class objects produced by msprime or SLiM, including
    subsampling, simplifying, and showing mutations on edges.
utils/src/exceptions/
    Raising and catching exceptions. Users can access the main
    exception class from the top level module: `toytree.ToyTreeError`
    and so do not need to access it here.
utils/src/networks/
    Network parsing functions. Currently `parse_network` is the only
    function, and is accessible at `toytree.utils.parse_network`.
utils/src/logging/
    The logger setup. Users can set the log level from the top level
    module with `toytree.set_log_level`, or turn it on/off with 
    `from loguru import logger; logger.disable('toytree')`.
utils/src/constants/
    A file with some constants that users should not need to access.
"""

from toytree.utils.src.exceptions import *
from toytree.utils.src.scrollable_canvas import ScrollableCanvas
from toytree.utils.src.browser import show
from toytree.utils.src.toytree_sequence import ToyTreeSequence as toytree_sequence
from toytree.utils.src.style_axes import (
    set_axes_ticks_external,
    set_axes_box_outline,
)



#!/usr/bin/env python

"""
The core subpackage contains modules for manipulating and drawing
ToyTree and MultiTree objects.

Examples
--------
>>> import toytree

Parse tree data to a ToyTree instance:
>>> tree = toytree.tree("((a,b),c);")
>>> tree = toytree.tree("https://eaton-lab.org/data/Cyathophora.tre")

Generate random ToyTrees
>>> tree = toytree.rtree.unittree(ntips=10)
>>> tree = toytree.rtree.bdtree(ntips=10, b=0.5, d=0.1)

Generate a MultiTree object
>>> trees = [toytree.rtree.unittree(10) for i in range(5)]
>>> mtree = toytree.mtree(trees)

"""

from toytree.core.tree import ToyTree
from toytree.core.node import Node
# from toytree.core.multitree import MultiTree

# easier acces to the main toyplot types
from toyplot.canvas import Canvas
from toyplot.coordinates import Cartesian
from toyplot.mark import Mark



#!/usr/bin/env python

"""Utility functions for annotations.
"""

# from toyplot.canvas import Canvas
from toyplot.coordinates import Cartesian
from toyplot.mark import Mark
from toytree import ToyTree
from toytree.drawing import ToyTreeMark
from toytree.utils import ToytreeError

CARTESIAN_TYPE_ERROR = """\
'axes' argument to annotation func must be a Cartesian object
containing a ToyTree Mark (drawing). For example:

>>> # get Cartesian axes from a toytree drawing
>>> canvas, axes, mark = tree.draw()
>>> tree.annotate.add_tip_markers(axes, size=9, color='red', xshift=20)
"""


def get_last_toytree_mark(axes: Cartesian) -> ToyTreeMark:
    """Return the last ToyTreeMark rendered on the Cartesian axes.

    """
    if not isinstance(axes, Cartesian):
        raise TypeError(CARTESIAN_TYPE_ERROR)
    targets = axes._scenegraph.targets(axes, "render")
    tree_marks = [i for i in targets if isinstance(i, ToyTreeMark)]
    if not tree_marks:
        raise ToytreeError(
            "No tree drawings (ToyTreeMark) have been rendered on the "
            "Cartesian axes. First draw a ToyTree.")
    return tree_marks[-1]


def assert_tree_matches_mark(tree: ToyTree, mark: Mark) -> None:
    """Raise exception if mark does not match up with tree data.
    """
    nnodes_in_mark = mark.ntable.shape[0]
    assert tree.nnodes == nnodes_in_mark, (
        f"Cannot annotate a tree drawing containing {tree.nnodes} nodes "
        f"with a different tree that has {nnodes_in_mark} nodes.")


if __name__ == "__main__":

    pass



#!/usr/bin/env python

"""Tree API allows calling submodule functions from a tree object.

For example, the function in the `toytree.mod` submodule can be called
as `toytree.mod.prune(tree)`, or the submodule can be called from a
ToyTree object as `tree.mod.prune()`. For some very common functions,
such as `root`, the function is also made available from the ToyTree
object directly without accessing the submodule, e.g., `tree.root()`.

To simplify the code and avoid errors in making these functions
accessible from three places we use a clever wrapper in functools and
a method described at the link below. This allows us to write the
function just once and to copy its docstring, signature, etc. to
automatically set it as a method in the API.

Thus, most function's source code is in the submodule folders, e.g.,
`toytree.mod.root`, and it is added as a method to the tree-level
subpackage API, e.g., `tree.mod` or to the ToyTree object itself, e.g.,
`tree.root` using the wrapper in this module.

References
----------
https://mgarod.medium.com/dynamically-add-a-method-to-a-class-in-python-c49204b85bd6
"""

from typing import TypeVar
from functools import wraps

ToyTree = TypeVar("ToyTree")
# Cartesian = TypeVar("Cartesian")


class SubPackageAPI:
    """API to acess methods from `toytree.mod` as `tree.mod.{function}`"""
    def __init__(self, tree: ToyTree):
        self._tree = tree


class TreeModAPI(SubPackageAPI):
    """API to acess methods from `toytree.mod` as `tree.mod.{function}`"""


class TreeDistanceAPI(SubPackageAPI):
    """API to acess methods from `toytree.distance` as `tree.distance.{function}`"""


class TreeEnumAPI(SubPackageAPI):
    """API to acess methods from `toytree.pcm` as `tree.enum.{function}`"""


class PhyloCompAPI(SubPackageAPI):
    """API to acess methods from `toytree.pcm` as `tree.pcm.{function}`"""


class AnnotationAPI(SubPackageAPI):
    """API to acess methods from `toytree.annotate` as `tree.annotate.{function}`"""


def add_subpackage_method(cls):
    """Clever approach to copy docstring and signature from func to API.

    Add this decorator to a function to make it accessible from an API

    Example
    -------
    >>> @add_subpackage_method(TreeModAPI)
    >>> def method(...):
    """
    def decorator(func):

        # creates a wrapper for a toytree.mod.{function} that copies
        # its docstring and arg signatures but sets self._tree as the
        # first argument of the function.
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return func(self._tree, *args, **kwargs)

        # sets: TreeModAPI.{function} = wrapper
        setattr(cls, func.__name__, wrapper)

        # Note we are not binding func, but wrapper which accepts self
        # but does exactly the same as func.
        # returning func means func can still be used normally
        return func
    return decorator


def add_toytree_method(cls):
    """Clever approach to copy docstring and signature from func to API.

    Add this decorator to a function to make it accessible from ToyTree

    Example
    -------
    >>> @add_toytree_method(ToyTree)
    >>> def method(...):
    """
    def decorator(func):

        # creates a wrapper for a toytree.mod.{function} that copies
        # its docstring and arg signatures but sets self._tree as the
        # first argument of the function.
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        # sets: TreeModAPI.{function} = wrapper
        setattr(cls, func.__name__, wrapper)

        # Note we are not binding func, but wrapper which accepts self
        # but does exactly the same as func.
        # returning func means func can still be used normally
        return func
    return decorator


if __name__ == "__main__":

    @add_toytree_method(ToyTree)
    @add_subpackage_method(TreeModAPI)
    def test(arg1: int, arg2: str) -> float:
        """This is a test."""
        return 32.0

    help(test)

    help(TreeModAPI.test)

    help(ToyTree.test)



#!/usr/bin/env python

"""Tree edge modification functions.

Tree edge modification functions set new 'dist' values to one or more
Nodes in a ToyTree.

Note
----
All 'edge' functions in `mod` are written so that the returned ToyTrees
are ready to be plotted. This means they update the cached heights
(Node._height attributes). These functions do not change the topology,
so idx values and x positions do not need to be updated.
"""

from typing import Dict, TypeVar, Optional
from loguru import logger
import numpy as np
from toytree import Node
from toytree.core.apis import TreeModAPI, add_subpackage_method

logger = logger.bind(name="toytree")
ToyTree = TypeVar("ToyTree")
Query = TypeVar("Query", str, int, Node)


__all__ = [
    "edges_scale_to_root_height",
    "edges_slider",
    "edges_multiplier",
    "edges_extend_tips_to_align",
    "edges_set_node_heights",
]


@add_subpackage_method(TreeModAPI)
def edges_scale_to_root_height(
    tree: ToyTree,
    treeheight: float = 1.,
    include_stem: bool = False,
    inplace: bool = False,
) -> ToyTree:
    """Return ToyTree rescaled to a specific total tree height.

    Edge lengths (Node dist values) are all multiplied by a constant
    factor to make the root Node height align at treeheight. By
    default tree is scaled to the root crown height, but stem height
    can also be specified.

    Parameters
    ----------
    treeheight: float
        Total tree height to scale tree to.
    include_stem: bool
        If True then the stem height is set instead of crown height.
    inplace: bool
        If True the tree is modified, else a copy is returned.

    Example
    -------
    >>> tree = toytree.rtree.unittree(10, seed=123)
    >>> tree = tree.mod.edges_scale_to_root_height(1000)
    >>> tree.draw(scale_bar=True);
    """
    tree = tree if inplace else tree.copy()

    # get total tree height
    height = (
        tree.treenode.height + tree.treenode.dist if include_stem
        else tree.treenode.height)

    # get multiplier
    ratio = treeheight / height

    # scale Nodes using cached Nodes, no ._update call necessary.
    for idx in range(tree.nnodes):
        tree[idx]._height *= ratio
        tree[idx]._dist *= ratio
    return tree


@add_subpackage_method(TreeModAPI)
def edges_slider(
    tree: ToyTree,
    # query: Query = None,
    prop: float = 0.999,
    seed: Optional[int] = None,
    inplace: bool = False,
    root: bool = False,
) -> ToyTree:
    """Return ToyTree with node heights randomly shifted within bounds.

    Node heights are moved up or down uniformly between their parent
    and highest child node heights in 'levelorder' (from root to tips).
    Root and tip heights are fixed, only internal node heights change.

    Parameters
    ----------
    prop: float
        The proportion or percentile of the edge bounds from which
        to sample new heights from.
    seed: int
        Random number generator seed used to sample new heights.
    inplace: bool
        Transform tree in place and return it, or return a copy.
    """
    tree = tree if inplace else tree.copy()
    rng = np.random.default_rng(seed)

    # smaller value means less jitter
    assert 0 < prop < 1, "prop must be a proportion >0 and < 1."

    # randomly sample the traversal direction
    order = rng.choice(["postorder", "preorder"])

    # traverse tree sampling new heights within constrained ranges
    for node in tree.traverse(order):

        # slide internal Nodes
        if (not node.is_root()) and (not node.is_leaf()):

            # the closest child to me
            min_child_dist = min([i.dist for i in node.children])

            # prop distance down toward child
            minh = node._height - (min_child_dist * prop)

            # prop towards parent
            maxh = node._height + (node.dist * prop)

            # node.height
            newheight = rng.uniform(minh, maxh)

            # how much lower am i?
            delta = newheight - node._height

            # edges from children to reach me
            for child in node.children:
                child._dist += delta

            # slide self to match
            node._dist -= delta
            node._height = newheight
    return tree


@add_subpackage_method(TreeModAPI)
def edges_multiplier(
    tree: ToyTree,
    alpha: float = 3.0,
    inplace: bool = False,
    seed: Optional[int] = None
) -> ToyTree:
    """Return ToyTree w/ all Nodes scaled by a random constant.

    This scales the entire tree height and all nodes in unison, similar
    to `edges_scale_to_root_height` but instead of setting the root
    to a specific value, the value is sampled from a gamma distribution.
    The current mean root height serves as the mean of the distribution
    mean = (a / b); the user enter's a value for a to determine the
    variance of the distribution (a / b**2), smaller values have larger
    variance.

    Parameters
    ----------
    alpha: float
        The node height multiplier will be sampled from a gamma dist
        with parameters (a,b), where b will be automatically set so
        that the mean = a/b = current root height.
    inplace: bool
        Transform tree in place and return it, or return a copy.
    seed: int or None or RNG
        Random number generator seed used to sample new heights.
    """
    tree = tree if inplace else tree.copy()
    beta = tree.treenode.height / alpha
    multiplier = np.random.default_rng(seed).gamma(shape=alpha, scale=beta)
    for node in tree:
        node._dist = node._dist * multiplier
        node._height = node._height * multiplier
    # Note: ToyTree._update call is not needed here.
    return tree


@add_subpackage_method(TreeModAPI)
def edges_extend_tips_to_align(
    tree: ToyTree,
    inplace: bool = False,
) -> ToyTree:
    """Return a ToyTree with tips Node dists extended to align.

    Tip Node dists are extended to align with the Node that is farthest
    from the root (defined as height=0). This is a simple way to make
    a tree appear ultrametric.

    See Also
    --------
    ... TODO: [alternative ultrametric scaling methods]

    Parameters
    ----------
    inplace: bool
        If True tree is modified in place, else a copy is
    """
    tree = tree if inplace else tree.copy()
    for node in tree[:tree.ntips]:
        node._dist += node._height
        node._height = 0
    return tree


@add_subpackage_method(TreeModAPI)
def edges_set_node_heights(
    tree: ToyTree,
    data: Dict[Query, float],
    inplace: bool = False,
) -> ToyTree:
    """Return a ToyTree with one or more Node heights set explicitly.

    Enter a dictionary mapping node idx to heights. Node idxs that
    are not included as keys will remain at there existing height, but
    their dist values may be changed to modify the heights of other
    Nodes, since height is an emergent property of the dist values of
    many connected Nodes.

    Parameters
    ----------
    data: Dict[Query, float]
        A dictionary mapping a Node query (Node, Node idx label, or Node
        name) to a float value for the new height of that Node.
    inplace: bool
        Tree is modified in place or returned as a copy.

    Examples
    --------
    >>> tre = toytree.rtree.unitree(10)
    >>> tre = tre.mod.edges_set_node_heights({10: 55, 11: 60, 12: 100})
    """
    # set node heights on a new tree copy
    tree = tree if inplace else tree.copy()

    # convert {query: float} to {idx: float} using Node int idx labels
    mapping = {
        tree.get_nodes(i)[0].idx: j for (i, j) in data.items()
    }

    # set node height to current value for those not in hdict
    for idx in range(tree.nnodes):
        if idx not in mapping:
            mapping[idx] = tree[idx]._height

    # iterate over nodes from tips to root
    for node in tree:  # .traverse("postorder"):
        # shorten or elongate child stems to reach node's new height
        node._height = mapping[node.idx]
        if node.up:
            node._dist = mapping[node.up.idx] - mapping[node.idx]

    # report warning if any edges end up negative
    for node in tree:
        if node._dist < 0:
            logger.warning(f"negative edge length {node._dist:.6g} @ Node {node._idx}")
    return tree


if __name__ == "__main__":

    import toytree
    TREE = toytree.rtree.rtree(ntips=6)
    TREE = toytree.rtree.unittree(ntips=6, treeheight=100, seed=123)

    NEW_HEIGHTS = {
        10: 200,
        9: 10,
        8: 150,
        7: 130,
        6: 80,
    }

    print(edges_scale_to_root_height(TREE, 100, False, False).get_node_data())
    print(edges_slider(TREE, 0.5).get_node_data())
    print(edges_multiplier(TREE, alpha=0.5).get_node_data())
    print(edges_extend_tips_to_align(TREE).get_node_data())
    print(edges_set_node_heights(TREE, NEW_HEIGHTS).get_node_data())



#!/usr/bin/env python

"""Custom exception classes.

We catch most errors that are of the ToytreeError BaseClass and try
to handle them more nicely than other random errors. This includes
adding a logging.error message.
"""

class ToytreeError(Exception):
    """BaseClass for many custom exceptions or common user errors."""


class ToytreeRegexError(Exception):
    """Exceptions for data getting/setting on Nodes."""
    pass


class NodeDataError(Exception):
    """Exceptions for data getting/setting on Nodes."""
    pass


class ToyColorError(ToytreeError):
    """Exceptions for parsing color inputs."""
    pass


class StyleSizeMismatchError(ToytreeError):
    pass


class StyleTypeMismatchError(ToytreeError):
    pass


class StyleColorMappingTupleError(ToytreeError):
    pass


class TreeNodeError(ToytreeError):
    """Exceptions for operating on immutable attributes."""
    pass


class NewickError(ToytreeError):
    pass


class NexusError(ToytreeError):
    pass


NODE_NOT_IN_TREE_ERROR = """\
"query Node belongs to a different tree (e.g., tree1) than the one
being queried (e.g., tree2). Make sure you are selecting the correct
Node, and consider using Node .idx or .name features as queries. Beware
.idx labels often change between trees. To avoid this warning try the
following to translate a query from one tree to another:

# get Node of interest from tree1 using your query
>>> node = tree1.get_mrca_node(*query)

# represent this Node by its set of leaf names
>>> tips = node.get_leaf_names()

# run your function of interest on tree2 using leaf names as query
>>> tree2.get_nodes(*tips)
"""

NODE_INDEXING_ERROR = """\
Error indexing from ToyTree object. You likely entered an invalid
Node query. Note: Nodes can be selected by their int idx label using
indexing, slicing, or as a list:
>>> tree[3]                         # Node(idx=3)
>>> tree[3:4]                       # [Node(idx=3), Node(idx=4)]
>>> tree[[3,10,2]]                  # [Node(idx=3), Node(idx=10), Node(idx=2)]

Also, see the `.get_nodes` function for more options to query Nodes
by int idx label, str name label, Node object, or a mixture:
>>> tree.get_nodes('name3', 0)        # [Node(idx=3), Node(idx=0)]
"""

NON_MONOPHYLETIC_OUTGROUP = """\
Cannot root on non-monophyletic outgroup ({})
If you want to root on the MRCA of these nodes try:
>>> mrca = tree.get_mrca_node(*query)
>>> tree.root(mrca)
"""

if __name__ == "__main__":

    # raise ValueError("HELLO WORLD")
    # raise ToytreeError("HELLO WORLD")
    import toytree
    tree = toytree.rtree.unittree(10)
    # tree.draw(node_colors=['red', 'blue'])
    raise ToytreeError("error ")
